# Copyright (c) 2018 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Manager
# (see https://github.com/MatthewScholefield/mycroft-light).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from contextlib import contextmanager


class MsmException(Exception):
    def __repr__(self):
        s = self.__str__().rstrip('\n')
        if '\n' in s:
            s = s.replace('\n', '\n\t') + '\n'
        return '{}({})'.format(self.__class__.__name__, s)


class GitException(MsmException):
    pass


class GitAuthException(GitException):
    def __repr__(self):
        return self.__class__.__name__


class MergeConflict(GitException):
    def __repr__(self):
        return self.__class__.__name__


class SkillModified(MsmException):
    """
    Raised when a skill cannot be updated because
    it has been modified by the user
    """
    pass


class RemoveException(MsmException):
    pass


class AlreadyRemoved(RemoveException):
    pass


class InstallException(MsmException):
    pass


class SkillNotFound(InstallException):
    pass


class SkillRequirementsException(InstallException):
    pass


class CloneException(InstallException):
    pass


class AlreadyInstalled(InstallException):
    pass


class NotInstalled(MsmException):
    pass


class SystemRequirementsException(InstallException):
    pass


class PipRequirementsException(InstallException):
    def __init__(self, code, stdout, stderr):
        self.code, self.stdout, self.stderr = code, stdout, stderr

    def __str__(self):
        return '\nPip returned code {}:\n{}\n{}'.format(
            self.code, self.stdout, self.stderr
        )


class MultipleSkillMatches(MsmException):
    def __init__(self, skills):
        self.skills = skills

    def __str__(self):
        return ', '.join(skill.name for skill in self.skills)


@contextmanager
def git_to_msm_exceptions():
    try:
        yield
    except Exception as e:
        msg = getattr(e, 'stderr', str(e)).replace('stderr:', '').strip()
        if 'Authentication failed for' in msg:
            raise GitAuthException(msg) from e
        if 'Not something we can merge' in msg or \
                'Not possible to fast-forward':
            raise MergeConflict(msg) from e
        raise GitException(msg) from e



# Copyright (c) 2018 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Manager
# (see https://github.com/MycroftAI/mycroft-skills-manager).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
import os
import yaml
from contextlib import contextmanager
from difflib import SequenceMatcher
from os.path import exists, join, basename
from threading import Lock
from typing import Callable

LOG = logging.getLogger(__name__)

# Branches which can be switched from when updating
# TODO Make this configurable
SWITCHABLE_BRANCHES = ['master']

# default constraints to use if no are given
DEFAULT_CONSTRAINTS = '/etc/mycroft/constraints.txt'
FIVE_MINUTES = 300


def _perform_pako_install(packages, system_packages=None):
    pass


@contextmanager
def work_dir(directory):
    old_dir = os.getcwd()
    os.chdir(directory)
    try:
        yield
    finally:
        os.chdir(old_dir)


def _backup_previous_version(func: Callable = None):
    return func


class SkillEntry:
    pip_lock = Lock()
    manifest_yml_format = {
        'dependencies': {
            'system': {},
            'exes': [],
            'skill': [],
            'python': []
        }
    }

    def __init__(self, name, path, url='', sha='', msm=None):
        url = url.rstrip('/')
        url = url[:-len('.git')] if url.endswith('.git') else url
        self.path = path
        self.url = url
        self.sha = sha
        self.msm = msm
        if msm:
            u = url.lower()
            self.meta_info = msm.repo.skills_meta_info.get(u, {})
        else:
            self.meta_info = {}
        if name is not None:
            self.name = name
        elif 'name' in self.meta_info:
            self.name = self.meta_info['name']
        else:
            self.name = basename(path)

        # TODO: Handle git:// urls as well
        from_github = False
        if url.startswith('https://'):
            url_tokens = url.rstrip("/").split("/")
            from_github = url_tokens[-3] == 'github.com' if url else False
        self.author = self.extract_author(url) if from_github else ''
        self.id = self.extract_repo_id(url) if from_github else self.name
        self.is_local = exists(path)
        self.old_path = None  # Path of previous version while upgrading

    @property
    def is_beta(self):
        return True

    @property
    def is_dirty(self):
        return False

    @property
    def skill_gid(self):
        """Format skill gid for the skill.

        This property does some Git gymnastics to determine its return value.
        When a device boots, each skill accesses this property several times.
        To reduce the amount of boot time, cache the value returned by this
        property.  Cache expires five minutes after it is generated.
        """
        LOG.debug('Generating skill_gid for ' + self.name)
        gid = ''
        if self.is_dirty:
            gid += '@|'
        if self.meta_info != {}:
            gid += self.meta_info['skill_gid']
        else:
            name = self.name.split('.')[0]
            gid += name
        return gid

    def __str__(self):
        return self.name

    def attach(self, remote_entry):
        """Attach a remote entry to a local entry"""
        return self

    @classmethod
    def from_folder(cls, path, msm=None, use_cache=True):
        """Find or create skill entry from folder path.

        Arguments:
            path:       path of skill folder
            msm:        mock_msm instance to use for caching and extended information
                        retrieval.
            use_cache:  Enable/Disable cache usage. defaults to True
        """
        if msm and use_cache:
            skills = {skill.path: skill for skill in msm.local_skills.values()}
            if path in skills:
                return skills[path]
        return cls(None, path, cls.find_git_url(path), msm=msm)

    @classmethod
    def create_path(cls, folder, url, name=''):
        return join(folder, '{}.{}'.format(
            name or cls.extract_repo_name(url), cls.extract_author(url)
        ).lower())

    @staticmethod
    def extract_repo_name(url):
        s = url.rstrip('/').split("/")[-1]
        a, b, c = s.rpartition('.git')
        if not c:
            return a
        return s

    @staticmethod
    def extract_author(url):
        return url.rstrip('/').split("/")[-2].split(':')[-1]

    @classmethod
    def extract_repo_id(cls, url):
        return '{}:{}'.format(cls.extract_author(url).lower(),
                              cls.extract_repo_name(url)).lower()

    @staticmethod
    def _tokenize(x):
        return x.replace('-', ' ').split()

    @staticmethod
    def _extract_tokens(s, tokens):
        s = s.lower().replace('-', ' ')
        extracted = []
        for token in tokens:
            extracted += [token] * s.count(token)
            s = s.replace(token, '')
        s = ' '.join(i for i in s.split(' ') if i)
        tokens = [i for i in s.split(' ') if i]
        return s, tokens, extracted

    @classmethod
    def _compare(cls, a, b):
        return SequenceMatcher(a=a, b=b).ratio()

    def match(self, query, author=None):
        search, search_tokens, search_common = self._extract_tokens(
            query, ['skill', 'fallback', 'mycroft']
        )

        name, name_tokens, name_common = self._extract_tokens(
            self.name, ['skill', 'fallback', 'mycroft']
        )

        weights = [
            (9, self._compare(name, search)),
            (9, self._compare(name.split(' '), search_tokens)),
            (2, self._compare(name_common, search_common)),
        ]
        if author:
            author_weight = self._compare(self.author, author)
            weights.append((5, author_weight))
            author_weight = author_weight
        else:
            author_weight = 1.0
        return author_weight * (
            sum(weight * val for weight, val in weights) /
            sum(weight for weight, val in weights)
        )

    def run_pip(self, constraints=None):
        return False

    def install_system_deps(self):
        return False

    def run_requirements_sh(self):
        return False

    def run_skill_requirements(self):
        return False

    def verify_info(self, info, fmt):
        pass

    def skill_info(self):
        yml_path = join(self.path, 'manifest.yml')
        if exists(yml_path):
            LOG.info('Reading from manifest.yml')
            with open(yml_path) as f:
                info = yaml.safe_load(f)
                self.verify_info(info, self.manifest_yml_format)
                return info or {}
        return {}

    def dependencies(self):
        return self.skill_info.get('dependencies') or {}

    def dependent_skills(self):
        skills = set()
        reqs = join(self.path, "skill_requirements.txt")
        if exists(reqs):
            with open(reqs, "r") as f:
                for i in f.readlines():
                    skill = i.strip()
                    if skill:
                        skills.add(skill)
        for i in self.dependencies.get('skill') or []:
            skills.add(i)
        return list(skills)

    def dependent_python_packages(self):
        reqs = join(self.path, "requirements.txt")
        req_lines = []
        if exists(reqs):
            with open(reqs, "r") as f:
                req_lines += f.readlines()
        req_lines += self.dependencies.get('python') or []
        # Strip comments
        req_lines = [line.split('#')[0].strip() for line in req_lines]
        return [line for line in req_lines if line]  # Strip empty lines

    def dependent_system_packages(self):
        return self.dependencies.get('system') or {}

    def remove(self):
        raise NotImplementedError

    def install(self, constraints=None):
        raise NotImplementedError

    def update_deps(self, constraints=None):
        raise NotImplementedError

    def _find_sha_branch(self):
        return ""

    def update(self):
        return False

    @staticmethod
    def find_git_url(path):
        return ''

    def __repr__(self):
        return '<SkillEntry {}>'.format(' '.join(
            '{}={}'.format(attr, self.__dict__[attr])
            for attr in ['name', 'author', 'is_local']
        ))



# Copyright (c) 2018 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Manager
# (see https://github.com/MatthewScholefield/mycroft-light).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Install, remove, update and track the skills on a device

MSM can be used on the command line but is also used by Mycroft core daemons.
"""
import logging
from glob import glob
from os import path
from typing import Dict, List

from xdg import BaseDirectory
from mock_msm.skill_entry import SkillEntry
from mock_msm.skill_repo import SkillRepo
from mock_msm.util import MsmProcessLock

LOG = logging.getLogger(__name__)

CURRENT_SKILLS_DATA_VERSION = 2
ONE_DAY = 86400


def save_device_skill_state(func):
    return func


class MycroftSkillsManager:
    SKILL_GROUPS = {'default', 'mycroft_mark_1', 'picroft', 'kde',
                    'respeaker', 'mycroft_mark_2', 'mycroft_mark_2pi'}

    def __init__(self, platform='default', old_skills_dir=None,
                 skills_dir=None, repo=None, versioned=True):
        self.platform = platform

        # Keep this variable alive for a while, is used to move skills from the
        # old config based location to XDG
        self.old_skills_dir = path.expanduser(old_skills_dir or '') or None
        self.skills_dir = (skills_dir or
                           BaseDirectory.save_data_path('mycroft/skills'))

        self.repo = repo or SkillRepo()
        self.versioned = versioned
        self.lock = MsmProcessLock()

        # Property placeholders
        self._all_skills = None
        self._default_skills = None
        self._local_skills = None
        self._device_skill_state = None

        self.saving_handled = False
        self.device_skill_state_hash = ''
        with self.lock:
            self._init_skills_data()

    def clear_cache(self):
        """Completely clear the skills cache."""
        pass

    @property
    def all_skills(self):
        if self._all_skills is None:
            self._all_skills = self._get_all_skills()
        return self._all_skills

    def _get_all_skills(self):
        remote_skills = self._get_remote_skills()
        all_skills = self._merge_remote_with_local(remote_skills)
        return all_skills

    def list(self):
        return self._get_all_skills()

    def _refresh_skill_repo(self):
        """Get the latest mycroft-skills repo code."""
        pass

    def _get_remote_skills(self):
        return {}

    def _merge_remote_with_local(self, remote_skills):
        """Merge the skills found in the repo with those installed locally."""
        all_skills = []

        for skill_file in glob(path.join(self.skills_dir, '*', '__init__.py')):
            skill = SkillEntry.from_folder(path.dirname(skill_file), msm=self,
                                           use_cache=False)
            all_skills.append(skill)

        return all_skills

    @property
    def local_skills(self):
        """Property containing a dictionary of local skills keyed by name."""
        if self._local_skills is None:
            self._local_skills = {
                s.name: s for s in self.all_skills if s.is_local
            }

        return self._local_skills

    @property
    def default_skills(self):
        return {}

    def list_all_defaults(self):  # type: () -> Dict[str, List[SkillEntry]]
        default_skills = {group: [] for group in self.SKILL_GROUPS}
        return default_skills

    def _init_skills_data(self):
        pass

    @property
    def device_skill_state(self):
        return {"mock": "data"}

    def _upgrade_skills_data(self):
        pass

    def _upgrade_to_v1(self):
        pass

    def _upgrade_to_v2(self):
        pass

    def _sync_device_skill_state(self):
        pass

    def _add_skills_to_state(self):
        pass

    def _remove_skills_from_state(self):
        pass

    def _update_skill_gid(self):
        pass

    def _determine_skill_origin(self, skill):
        if skill.name in self.default_skills:
            origin = 'default'
        elif skill.url:
            origin = 'cli'
        else:
            origin = 'non-msm'
        return origin

    def write_device_skill_state(self, data=None):
        pass

    def install(self, param, author=None, constraints=None, origin=''):
        raise NotImplementedError

    def remove(self, param, author=None):
        pass

    def update_all(self):
        pass

    def update(self, skill=None, author=None):
        pass

    def apply(self, func, skills, max_threads=20):
        return None

    def install_defaults(self):
        pass

    def _invalidate_skills_cache(self, new_value=None):
        pass

    def find_skill(self, param, author=None, skills=None):
        return None



from setuptools import setup


setup(
    name='mock_msm',
    version='0.9.1',
    packages=['mock_msm'],
    url='https://github.com/MycroftAI/mycroft-skills-manager',
    license='Apache-2.0',
    author='jarbasAI',
    author_email='jarbasai@mailfence.com',
    description='Mock Mycroft Skills Manager',
    long_description_content_type="text/markdown",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ]
)



# Copyright (c) 2018 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Manager
# (see https://github.com/MatthewScholefield/mycroft-light).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from glob import glob
from os.path import join, basename
from xdg import BaseDirectory
import logging

LOG = logging.getLogger(__name__)

MYCROFT_SKILLS_DATA = ("https://raw.githubusercontent.com/"
                       "MycroftAI/mycroft-skills-data")
FIVE_MINUTES = 300


def download_skills_data(branch, path):
    return {}


def load_cached_skills_data(path):
    return {}


def load_skills_data(branch, path):
    return {}


class SkillRepo:
    def __init__(self, url=None, branch=None):
        self.path = join(BaseDirectory.save_data_path('mycroft'),
                         'skills-repo')
        self.url = url or "https://github.com/MycroftAI/mycroft-skills"
        self.branch = branch or "20.08"
        self.repo_info = {}

    @property
    def skills_meta_info(self):
        return {}

    def read_file(self, filename):
        with open(join(self.path, filename)) as f:
            return f.read()

    def __prepare_repo(self):
        pass

    def update(self):
        pass

    def get_skill_data(self):
        """ generates tuples of name, path, url, sha """
        for name, path, url, sha in []:
            yield name, path, url, sha

    def get_shas(self):
        for folder, sha in []:
            yield folder, sha

    def get_default_skill_names(self):
        for defaults_file in glob(join(self.path, 'DEFAULT-SKILLS*')):
            with open(defaults_file) as f:
                skills = list(filter(
                    lambda x: x and not x.startswith('#'),
                    map(str.strip, f.read().split('\n'))
                ))
            platform = basename(defaults_file).replace('DEFAULT-SKILLS', '')
            platform = platform.replace('.', '') or 'default'
            yield platform, skills



"""Functions related to manipulating the skills.json file."""
import json
from logging import getLogger
from os.path import join

from xdg import BaseDirectory

LOG = getLogger(__name__)


def get_state_path():
    """Get complete path for skill state file.

    Returns:
        (str) path to skills.json
    """
    return join(BaseDirectory.save_data_path('mycroft'), 'skills.json')


def load_device_skill_state() -> dict:
    return {}


def write_device_skill_state(data: dict):
    pass


def get_skill_state(name, device_skill_state) -> dict:
    return {}


def initialize_skill_state(name, origin, beta, skill_gid) -> dict:
    """Create a new skill entry

    Arguments:
        name: skill name
        origin: the source of the installation
        beta: Boolean indicating wether the skill is in beta
        skill_gid: skill global id
    Returns:
        populated skills entry
    """
    return dict(
        name=name,
        origin=origin,
        beta=beta,
        status='active',
        installed=0,
        updated=0,
        installation='installed',
        skill_gid=skill_gid
    )


def device_skill_state_hash(data):
    return hash(json.dumps(data, sort_keys=True))
